#ifndef MISC_FUNCTIONS
#define MISC_FUNCTIONS

#include <exec/types.h>
#include <graphics/text.h>
#include <dos/dos.h>


ULONG ReadLine(BPTR file, char *string, ULONG max=1e6);
ULONG WriteLine(BPTR file, char *string);


inline LONG TextFontLength(TextFont *font, const char *str, int len);
inline LONG TextFontHeight(TextFont *font);
inline LONG TextFontBase(TextFont *font);

inline BOOL IsEmptyList(List *list);
inline Node *ListHead(List *list);
inline Node *ListTail(List *list);
inline Node *NextNode(Node *node);
inline Node *PrevNode(Node *node);
inline void TraverseMinList(List *list, void (*visitnode)(Node*));

inline int RawFmt(char *recieve, char *format, ...);
inline int RawFmtQuick(char *recieve, char *format, ...);

inline APTR PooledAllocVec(void *pool, ULONG size);
inline void PooledFreeVec(void *pool, APTR mem);


#include <pragma/exec_lib.h>
#include <exec/lists.h>
#include <exec/nodes.h>
#include <pragma/graphics_lib.h>
#include <graphics/GfxBase.h>
#include <stdarg.h>
#include <pragma/dos_lib.h>
#include <dos/dostags.h>
#include <string.h>
#include <ClassLibrary/SemaphoreC.h>




//{ "TextFontLength"
inline LONG TextFontLength(TextFont *font, const char *str, int len)
{
	extern struct GfxBase *GfxBase;
	int lpc;
	LONG width = 0;
	int currch;

	if(!(font))
	{
		font = ((struct GfxBase *)GfxBase)->DefaultFont;
	}

	if(font->tf_Flags & FPF_PROPORTIONAL)
	{
		for(lpc=0; lpc<len; lpc++)
		{
			currch = str[lpc] - font->tf_LoChar;
			width += ((WORD *)font->tf_CharSpace)[currch];
			width += ((WORD *)font->tf_CharKern)[currch];
		}
	}
	else
	{
		width = font->tf_XSize * len;
	}

	return width;
}
//}

//{ "TextFontHeight"
inline LONG TextFontHeight(TextFont *font)
{
	extern struct GfxBase *GfxBase;
	if(!(font)) font = ((struct GfxBase *)GfxBase)->DefaultFont;
	return font->tf_YSize;
}
//}

//{ "TextFontBase"
inline LONG TextFontBase(TextFont *font)
{
	extern struct GfxBase *GfxBase;
	if(!(font)) font = ((struct GfxBase *)GfxBase)->DefaultFont;
	return font->tf_Baseline;
}
//}

//{ "list functions"
inline BOOL IsEmptyList(List *list)
{
	return(NULL == list->lh_Head->ln_Succ);
}


inline Node *ListHead(List *list)
{
	if(NULL != list->lh_Head->ln_Succ)
		return list->lh_Head;
	return NULL;
}


inline Node *ListTail(List *list)
{
	if(NULL != list->lh_Head->ln_Succ)
		return list->lh_TailPred;
	return NULL;
}

inline Node *NextNode(Node *node)
{
	if(node->ln_Succ->ln_Succ != NULL)
		return node->ln_Succ;
	return NULL;
}


inline Node *PrevNode(Node *node)
{
	if(node->ln_Pred->ln_Pred != NULL)
		return node->ln_Pred;
	return NULL;
}


inline void TraverseMinList(List *list, void (*visitnode)(Node*))
{
	Node *node = ListHead(list);
	while(node = NextNode(node))
	{
		visitnode(node);
	}
}
//}

//{ "RawFmt"
static char *__putchbuffer__;

static void putch(register __d0 char ch)
{
	*__putchbuffer__++ = ch;
}

inline int RawFmt(char *recieve, char *format, ...)
{
	extern SemaphoreC RawFmtSem;
	RawFmtSem.Obtain();
	__putchbuffer__ = recieve;
	RawDoFmt(format, (APTR)((unsigned int)(&format)+sizeof(format)), (void(*)())putch, recieve);
	int result = __putchbuffer__ - recieve - 1;
	RawFmtSem.Release();
	return result;
}

inline int RawFmtQuick(char *recieve, char *format, ...)
{
	__putchbuffer__ = recieve;
	RawDoFmt(format, (APTR)((unsigned int)(&format)+sizeof(format)), (void(*)())putch, recieve);
	return __putchbuffer__ - recieve - 1;
}
//}

//{ "pooled"
inline APTR PooledAllocVec(void *pool, ULONG size)
{
	size+=4;
	ULONG *ptr = (ULONG *)AllocPooled(pool, size);
	*ptr = size;
	ptr++;
	return (APTR)ptr;
}

inline void PooledFreeVec(void *pool, APTR mem)
{
	ULONG size = *(--(ULONG *)mem);
	FreePooled(pool, mem, size);
}
//}


#endif
